# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""StudyConfig used by pythia policies."""

import enum
import attr

from vizier._src.pyvizier.shared import base_study_config


class StudyState(enum.Enum):
  ACTIVE = 'ACTIVE'
  ABORTED = 'ABORTED'
  COMPLETED = 'COMPLETED'


@attr.define
class StudyStateInfo:
  state: StudyState = attr.field(
      converter=StudyState, validator=attr.validators.instance_of(StudyState))
  explanation: str = attr.field(default='')


@attr.define(frozen=True, init=True)
class StudyDescriptor:
  """Light-weight, cross-platform summary of Study."""

  config: base_study_config.ProblemStatement = attr.ib(
      init=True,
      validator=[
          attr.validators.optional(
              attr.validators.instance_of(base_study_config.ProblemStatement))
      ])

  guid: str = attr.ib(
      init=True,
      validator=[attr.validators.optional(attr.validators.instance_of(str))],
      kw_only=True)

  max_trial_id: int = attr.ib(
      init=True,
      validator=[attr.validators.optional(attr.validators.instance_of(int))],
      kw_only=True)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""PSD Kernel for masking out dimensions."""

import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfpk = tfp.math.psd_kernels
tfpke = tfp.experimental.psd_kernels


class MaskFeatures(tfpk.PositiveSemidefiniteKernel):
  """Kernel for masking out features by setting them to zero.

  By masking out feature dimensions to zero (or to any other finite value), we
  ensure that when the Stochastic Process Model aware of this masking ingests
  these features (which may be NaN), we will get finite gradients. See
  https://github.com/tensorflow/probability/blob/main/discussion/where-nan.pdf
  for more details.
  """

  def __init__(
      self,
      kernel: tfpk.PositiveSemidefiniteKernel,
      dimension_is_missing: tfpke.ContinuousAndCategoricalValues,
  ):
    parameters = dict(locals())
    self._kernel = kernel
    self._dimension_is_missing = dimension_is_missing

    def _transformation_fn(x, *_):
      return jax.tree_util.tree_map(
          lambda x_, d: jnp.where(d, jnp.zeros_like(x_), x_),
          x,
          dimension_is_missing,
      )

    self._mask_kernel = tfpk.FeatureTransformed(kernel, _transformation_fn)

    super(MaskFeatures, self).__init__(
        feature_ndims=kernel.feature_ndims,
        dtype=kernel.dtype,
        name='MaskFeatures',
        validate_args=False,
        parameters=parameters,
    )

  @property
  def kernel(self):
    return self._kernel

  @property
  def dimension_is_missing(self):
    return self._dimension_is_missing

  def _apply(self, x1, x2, example_ndims=0):
    return self._mask_kernel._apply(x1, x2, example_ndims)  # pylint: disable=protected-access

  def _matrix(self, x1, x2):
    return self._mask_kernel._matrix(x1, x2)  # pylint: disable=protected-access

  @classmethod
  def _parameter_properties(cls, dtype):
    return dict(
        kernel=tfp.internal.parameter_properties.BatchedComponentProperties(),
        dimension_is_missing=(
            tfp.internal.parameter_properties.ParameterProperties(
                event_ndims=lambda self: self.kernel.feature_ndims
            )
        ),
    )


def _mask_features_flatten(v):
  children = (v.kernel, v.dimension_is_missing)
  return (children, None)


def _mask_features_unflatten(_, children):
  return MaskFeatures(*children)


jax.tree_util.register_pytree_node(
    MaskFeatures, _mask_features_flatten, _mask_features_unflatten
)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Abstraction for Vizier Tuner."""

import abc
import threading
from typing import Dict, NewType, Optional, Union

import attrs
import pyglove as pg
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.pyglove import converters
from vizier._src.pyglove import pythia as pg_pythia
from vizier.client import client_abc


# Pyglove-Vizier interactions only accept ExpandedStudyName. It is distinguished
# from the name that is passed into Backend.__init__ which does not include
# prefix.
ExpandedStudyName = NewType('ExpandedStudyName', str)


@attrs.define(eq=True, frozen=True)
class StudyKey:
  """Simple immutable structure used for looking up policies in a dict."""
  _owner: str
  _name: ExpandedStudyName


PolicyCache = Dict[StudyKey, pg_pythia.TunerPolicy]


class VizierTuner(abc.ABC):
  """Abstraction for Vizier tuner.

  Each platform (e.g. Google, OSS, Vertex) should have their own implementation
  of this abstraction.
  """
  # NOTE(chansoo): All of the methods can technically become a classmethod.

  def __init__(self):
    self._pythia_lock = threading.Lock()

  @abc.abstractmethod
  def get_tuner_id(self, algorithm: pg.DNAGenerator) -> str:
    """Get identifier of this tuner instance."""

  def start_pythia_service(self, policy_cache: PolicyCache) -> None:
    """Start pythia service on current machine."""
    with self._pythia_lock:
      self._start_pythia_service(policy_cache)

  @abc.abstractmethod
  def _start_pythia_service(
      self, policy_cache: dict[StudyKey, pg_pythia.TunerPolicy]
  ) -> None:
    """Starts pythia service _only if_ there is no pythia service running.

    Args:
      policy_cache: The pythia service will have the reference to policies
        created by the backend. It is expected to use the study keys to find
        the policies, instead of creating new ones from scratch.
    """

  @abc.abstractmethod
  def load_prior_study(self, resource_name: str) -> client_abc.StudyInterface:
    """Loads prior study identified by the resource name."""

  @abc.abstractmethod
  def create_study(
      self,
      problem: vz.ProblemStatement,
      converter: converters.VizierConverter,
      owner: str,
      name: str,
      algorithm: pg.DNAGenerator,
      stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy] = None,
  ) -> client_abc.StudyInterface:
    """Creates a new study."""

  @abc.abstractmethod
  def get_group_id(self, group_id: Union[None, int, str] = None) -> str:
    """Get worker ID."""

  @abc.abstractmethod
  def ping_tuner(self, tuner_id: str) -> bool:
    """See if the tuner is alive."""

  @abc.abstractmethod
  def pythia_supporter(
      self, study: client_abc.StudyInterface
  ) -> pythia.PolicySupporter:
    """Creates a pythia policy supporter for this study."""

  @abc.abstractmethod
  def use_pythia_for_study(self, study: client_abc.StudyInterface) -> None:
    """Uses current Pythia service to serve the input study."""

  @classmethod
  @abc.abstractmethod
  def load_study(
      cls, owner: str, name: ExpandedStudyName
  ) -> client_abc.StudyInterface:
    """Loads a study identified by (owner, name) pair."""



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Service-related policy factories."""

# pylint:disable=g-import-not-at-top
import functools
import time

from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.policies import designer_policy as dp


class DefaultPolicyFactory(pythia.PolicyFactory):
  """Default Policy Factory used in Pythia service."""

  def __call__(
      self,
      problem_statement: vz.ProblemStatement,
      algorithm: str,
      policy_supporter: pythia.PolicySupporter,
      study_name: str,
  ) -> pythia.Policy:
    del study_name
    # TODO: Compare string to Algorithm enums.
    if algorithm in (
        'DEFAULT',
        'ALGORITHM_UNSPECIFIED',
        'GP_UCB_PE',
    ):
      from vizier._src.algorithms.designers import gp_ucb_pe

      return dp.DesignerPolicy(policy_supporter, gp_ucb_pe.VizierGPUCBPEBandit)
    elif algorithm == 'GAUSSIAN_PROCESS_BANDIT':
      from vizier._src.algorithms.designers import gp_bandit

      return dp.DesignerPolicy(
          policy_supporter, gp_bandit.VizierGPBandit.from_problem
      )
    elif algorithm == 'RANDOM_SEARCH':
      from vizier._src.algorithms.policies import random_policy

      return random_policy.RandomPolicy(policy_supporter)
    elif algorithm == 'QUASI_RANDOM_SEARCH':
      from vizier._src.algorithms.designers import quasi_random

      return dp.PartiallySerializableDesignerPolicy(
          problem_statement,
          policy_supporter,
          quasi_random.QuasiRandomDesigner.from_problem,
      )
    elif algorithm == 'GRID_SEARCH':
      from vizier._src.algorithms.designers import grid

      return dp.PartiallySerializableDesignerPolicy(
          problem_statement,
          policy_supporter,
          grid.GridSearchDesigner.from_problem,
      )
    elif algorithm == 'SHUFFLED_GRID_SEARCH':
      from vizier._src.algorithms.designers import grid

      shuffle_seed = int(time.time())
      grid_factory = functools.partial(
          grid.GridSearchDesigner.from_problem, shuffle_seed=shuffle_seed
      )
      return dp.PartiallySerializableDesignerPolicy(
          problem_statement,
          policy_supporter,
          grid_factory,
      )
    elif algorithm == 'NSGA2':
      from vizier._src.algorithms.evolution import nsga2

      return dp.PartiallySerializableDesignerPolicy(
          problem_statement, policy_supporter, nsga2.NSGA2Designer
      )
    elif algorithm == 'EMUKIT_GP_EI':
      from vizier._src.algorithms.designers import emukit

      return dp.DesignerPolicy(policy_supporter, emukit.EmukitDesigner)
    elif algorithm == 'BOCS':
      from vizier._src.algorithms.designers import bocs

      return dp.DesignerPolicy(policy_supporter, bocs.BOCSDesigner)
    elif algorithm == 'HARMONICA':
      from vizier._src.algorithms.designers import harmonica

      return dp.DesignerPolicy(policy_supporter, harmonica.HarmonicaDesigner)
    elif algorithm == 'CMA_ES':
      from vizier._src.algorithms.designers import cmaes

      return dp.PartiallySerializableDesignerPolicy(
          problem_statement, policy_supporter, cmaes.CMAESDesigner
      )
    elif algorithm == 'EAGLE_STRATEGY':
      from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy

      return dp.PartiallySerializableDesignerPolicy(
          problem_statement,
          policy_supporter,
          eagle_strategy.EagleStrategyDesigner,
      )
    else:
      raise ValueError(f'Algorithm {algorithm} is not registered.')



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for gp_ucb_pe."""

import copy
from typing import Any, Tuple

import jax
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.core import abstractions
from vizier._src.algorithms.designers import gp_ucb_pe
from vizier._src.algorithms.optimizers import eagle_strategy as es
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier.jax import optimizers
from vizier.pyvizier.converters import padding
from vizier.testing import test_studies

from absl.testing import absltest
from absl.testing import parameterized

ensemble_ard_optimizer = optimizers.default_optimizer()


def _extract_predictions(
    metadata: Any,
) -> Tuple[float, float, float, float, bool]:
  pred = metadata.ns('prediction_in_warped_y_space')
  return (
      float(pred['mean']),
      float(pred['stddev']),
      float(pred['stddev_from_all']),
      float(pred['acquisition']),
      bool(pred['use_ucb'] == 'True'),
  )


class GpUcbPeTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(iters=3, batch_size=5, num_seed_trials=5),
      dict(iters=5, batch_size=1, num_seed_trials=2),
      dict(iters=5, batch_size=3, num_seed_trials=2, ensemble_size=3),
      dict(iters=3, batch_size=5, num_seed_trials=5, applies_padding=True),
      dict(iters=5, batch_size=1, num_seed_trials=2, pe_overwrite=True),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          applies_padding=True,
          optimize_set_acquisition_for_exploration=True,
      ),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          applies_padding=True,
          optimize_set_acquisition_for_exploration=True,
          search_space=test_studies.flat_categorical_space(),
      ),
  )
  def test_on_flat_space(
      self,
      iters: int = 5,
      batch_size: int = 1,
      num_seed_trials: int = 1,
      ard_optimizer: str = 'default',
      ensemble_size: int = 1,
      applies_padding: bool = False,
      pe_overwrite: bool = False,
      optimize_set_acquisition_for_exploration: bool = False,
      search_space: vz.SearchSpace = (
          test_studies.flat_continuous_space_with_scaling()
      ),
  ):
    # We use string names so that test case names are readable. Convert them
    # to objects.
    if ard_optimizer == 'default':
      ard_optimizer = optimizers.default_optimizer()
    problem = vz.ProblemStatement(search_space)
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    vectorized_optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=es.VectorizedEagleStrategyFactory(),
        max_evaluations=100,
    )
    designer = gp_ucb_pe.VizierGPUCBPEBandit(
        problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        num_seed_trials=num_seed_trials,
        ard_optimizer=ard_optimizer,
        metadata_ns='gp_ucb_pe_bandit_test',
        config=gp_ucb_pe.UCBPEConfig(
            ucb_coefficient=10.0,
            explore_region_ucb_coefficient=0.5,
            cb_violation_penalty_coefficient=10.0,
            ucb_overwrite_probability=0.0,
            pe_overwrite_probability=1.0 if pe_overwrite else 0.0,
            optimize_set_acquisition_for_exploration=(
                optimize_set_acquisition_for_exploration
            ),
        ),
        ensemble_size=ensemble_size,
        padding_schedule=padding.PaddingSchedule(
            num_trials=padding.PaddingType.MULTIPLES_OF_10
            if applies_padding
            else padding.PaddingType.NONE,
        ),
        rng=jax.random.PRNGKey(1),
    )

    all_active_trials = []
    all_trials = []
    trial_id = 1
    # Simulates batch suggestions with delayed feedback: the first two batches
    # are generated by the designer without any completed trials (but all with
    # active trials). Starting from the third batch, the oldest batch gets
    # completed and updated to the new designer with all the active trials, and
    # the designer then makes a new batch of suggestions. The last two batches
    # of suggestions are again made with only active trials being updated to
    # the designer.
    for idx in range(iters + 2):
      suggestions = designer.suggest(batch_size)
      self.assertLen(suggestions, batch_size)
      for suggestion in suggestions:
        problem.search_space.assert_contains(suggestion.parameters)
        all_active_trials.append(suggestion.to_trial(trial_id))
        all_trials.append(copy.deepcopy(all_active_trials[-1]))
        trial_id += 1
      completed_trials = []
      # Starting from the second until the last but two batch, complete the
      # oldest batch of suggestions.
      if idx > 0 and idx < iters:
        for _ in range(batch_size):
          measurement = vz.Measurement()
          for mi in problem.metric_information:
            measurement.metrics[mi.name] = float(
                jax.random.uniform(
                    jax.random.PRNGKey(1),
                    minval=mi.min_value_or(lambda: -10.0),
                    maxval=mi.max_value_or(lambda: 10.0),
                )
            )
          completed_trials.append(
              all_active_trials.pop(0).complete(measurement)
          )
      designer.update(
          completed=abstractions.CompletedTrials(completed_trials),
          all_active=abstractions.ActiveTrials(all_active_trials),
      )

    self.assertLen(all_trials, (iters + 2) * batch_size)

    # The suggestions after the seeds up to the first two batches are expected
    # to be generated by the PE acquisition function.
    for jdx in range(2 * batch_size):
      # Before the designer was updated with enough trials, the suggested
      # batches were seeds, not from acquisition optimization.
      if (jdx // batch_size) * batch_size >= num_seed_trials:
        _, _, _, acq, use_ucb = _extract_predictions(
            all_trials[jdx].metadata.ns('gp_ucb_pe_bandit_test')
        )
        self.assertFalse(use_ucb)
        if not optimize_set_acquisition_for_exploration:
          self.assertGreaterEqual(acq, 0.0, msg=f'suggestion: {jdx}')

    for idx in range(2, iters + 2):
      # Skips seed trials, which are not generated by acquisition function
      # optimization.
      if idx * batch_size < num_seed_trials:
        continue
      set_acq_value = None
      stddev_from_all_list = []
      for jdx in range(batch_size):
        mean, stddev, stddev_from_all, acq, use_ucb = _extract_predictions(
            all_trials[idx * batch_size + jdx].metadata.ns(
                'gp_ucb_pe_bandit_test'
            )
        )
        if jdx == 0 and idx < (iters + 1) and not pe_overwrite:
          # Except for the last batch of suggestions, the acquisition value of
          # the first suggestion in a batch is expected to be UCB, which
          # combines the predicted mean based only on completed trials and the
          # predicted standard deviation based on all trials.
          self.assertAlmostEqual(mean + 10.0 * stddev_from_all, acq)
          self.assertTrue(use_ucb)
          continue

        self.assertFalse(use_ucb)
        if optimize_set_acquisition_for_exploration:
          stddev_from_all_list.append(stddev_from_all)
          if set_acq_value is None:
            set_acq_value = acq
          else:
            self.assertAlmostEqual(set_acq_value, acq)
        else:
          # Because `ucb_overwrite_probability` is set to 0.0, when the designer
          # makes suggestions without seeing newer completed trials, it uses the
          # Pure-Exploration acquisition function. In this test, that happens
          # on the entire last batch and the second until the last suggestions
          # in every batch. The Pure-Exploration acquisition values are standard
          # deviation predictions based on all trials (completed and pending),
          # and are expected to be not much larger than the standard deviation
          # predictions based only on completed trials.
          self.assertLessEqual(
              acq, 2 * stddev, msg=f'batch: {idx}, suggestion: {jdx}'
          )
      if optimize_set_acquisition_for_exploration:
        geometric_mean_of_pred_cov_eigs = np.exp(
            set_acq_value / (batch_size - 1)
        )
        arithmetic_mean_of_pred_cov_eigs = np.mean(
            np.square(stddev_from_all_list)
        )
        self.assertLessEqual(
            geometric_mean_of_pred_cov_eigs, arithmetic_mean_of_pred_cov_eigs
        )

  def test_ucb_overwrite(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    vectorized_optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=es.VectorizedEagleStrategyFactory(),
        max_evaluations=100,
    )
    designer = gp_ucb_pe.VizierGPUCBPEBandit(
        problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        metadata_ns='gp_ucb_pe_bandit_test',
        num_seed_trials=1,
        config=gp_ucb_pe.UCBPEConfig(
            ucb_coefficient=10.0,
            explore_region_ucb_coefficient=0.5,
            cb_violation_penalty_coefficient=10.0,
            ucb_overwrite_probability=1.0,
        ),
        padding_schedule=padding.PaddingSchedule(
            num_trials=padding.PaddingType.MULTIPLES_OF_10
        ),
        rng=jax.random.PRNGKey(1),
    )

    trial_id = 1
    batch_size = 5
    iters = 3
    rng = jax.random.PRNGKey(1)
    all_trials = []
    # Simulates a batch suggestion loop that completes a full batch of
    # suggestions before asking for the next batch.
    for _ in range(iters):
      suggestions = designer.suggest(count=batch_size)
      self.assertLen(suggestions, batch_size)
      completed_trials = []
      for suggestion in suggestions:
        problem.search_space.assert_contains(suggestion.parameters)
        trial_id += 1
        measurement = vz.Measurement()
        for mi in problem.metric_information:
          measurement.metrics[mi.name] = float(
              jax.random.uniform(
                  rng,
                  minval=mi.min_value_or(lambda: -10.0),
                  maxval=mi.max_value_or(lambda: 10.0),
              )
          )
          rng, _ = jax.random.split(rng)
        completed_trials.append(
            suggestion.to_trial(trial_id).complete(measurement)
        )
      all_trials.extend(completed_trials)
      designer.update(
          completed=abstractions.CompletedTrials(completed_trials),
          all_active=abstractions.ActiveTrials(),
      )

    self.assertLen(all_trials, iters * batch_size)

    for idx, trial in enumerate(all_trials):
      if idx < batch_size:
        # Skips the first batch of suggestions, which are generated by the
        # seeding designer, not acquisition function optimization.
        continue
      # Because `ucb_overwrite_probability` is 1, all suggestions after the
      # first batch are expected to be generated by UCB. Within a batch, the
      # first suggestion's UCB value is expected to use predicted standard
      # deviation based only on completed trials, while the UCB values of
      # the second to the last suggestions are expected to use the predicted
      # standard deviations based on completed and active trials.
      mean, stddev, stddev_from_all, acq, use_ucb = _extract_predictions(
          trial.metadata.ns('gp_ucb_pe_bandit_test')
      )
      self.assertAlmostEqual(
          mean + 10.0 * (stddev_from_all if idx % batch_size > 0 else stddev),
          acq,
      )
      self.assertTrue(use_ucb)


if __name__ == '__main__':
  jax.config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Analyzers for BenchmarkStates for fast comparisons and statistics."""

import collections
import json
from typing import Callable, Dict, Optional, Sequence, Tuple

from absl import logging
import attrs
import numpy as np
import pandas as pd
from vizier import benchmarks
from vizier import pyvizier as vz
from vizier._src.benchmarks.analyzers import convergence_curve
from vizier.benchmarks import experimenters


RECORD_OBJECTIVE_KEY = 'objective'


@attrs.define(init=True, kw_only=True)
class PlotElement:
  """PlotElement with relevant information for a subplot."""

  curve: Optional[convergence_curve.ConvergenceCurve] = attrs.field(
      default=None,
      validator=attrs.validators.optional(
          attrs.validators.instance_of(convergence_curve.ConvergenceCurve)
      ),
  )
  plot_array: Optional[np.ndarray] = attrs.field(
      default=None,
      validator=attrs.validators.optional(
          attrs.validators.instance_of(np.ndarray)
      ),
  )
  # Error-bar uses curve, whereas histogram/scatter uses plot_array.
  plot_type: str = attrs.field(
      default='error-bar',
      validator=attrs.validators.in_(['error-bar', 'histogram', 'scatter']),
  )
  xlabel: str = attrs.field(
      default='Num Trials', validator=attrs.validators.instance_of(str)
  )
  yscale: str = attrs.field(
      default='linear',
      validator=attrs.validators.in_(['linear', 'symlog', 'logit']),
  )
  # Lower and upper percentiles to display for error bar.
  percentile_error_bar: Tuple[int, int] = attrs.field(
      default=(25, 75),
      validator=attrs.validators.deep_iterable(
          member_validator=attrs.validators.instance_of(int),
          iterable_validator=attrs.validators.instance_of(tuple),
      ),
  )


# Stores all relevant information and plots for a specific BenchmarkState.
@attrs.define(init=True, kw_only=True)
class BenchmarkRecord:
  algorithm: str = attrs.field(
      default='',
      validator=attrs.validators.instance_of(str),
  )
  experimenter_metadata: vz.Metadata = attrs.field(
      factory=vz.Metadata, validator=attrs.validators.instance_of(vz.Metadata)
  )
  plot_elements: Dict[str, PlotElement] = attrs.field(factory=dict)


class BenchmarkStateAnalyzer:
  """Analyzer for BenchmarkStates."""

  @classmethod
  def to_curve(
      cls,
      states: list[benchmarks.BenchmarkState],
      flip_signs_for_min: bool = False,
  ) -> convergence_curve.ConvergenceCurve:
    """Generates a ConvergenceCurve from a batch of BenchmarkStates.

    Each state in batch should represent the same study (different repeat).

    Args:
      states: List of BenchmarkStates.
      flip_signs_for_min: If true, flip signs of curve when it is MINIMIZE
        metric.

    Returns:
      Convergence curve with batch size equal to length of states.

    Raises:
      ValueError: When problem statements are not the same or is multiobjective.
    """
    if not states:
      raise ValueError('Empty States.')

    problem_statement = states[0].experimenter.problem_statement()

    curves = []
    for state in states:
      if problem_statement != state.experimenter.problem_statement():
        raise ValueError(
            f'States must have same problem {problem_statement}'
            f' and {state.experimenter.problem_statement()}'
        )
      state_trials = state.algorithm.supporter.GetTrials()

      converter = (
          convergence_curve.MultiMetricCurveConverter.from_metrics_config(
              problem_statement.metric_information,
              flip_signs_for_min=flip_signs_for_min,
          )
      )
      curve = converter.convert(state_trials)
      curves.append(curve)
    return convergence_curve.ConvergenceCurve.align_xs(curves)[0]

  @classmethod
  def to_record(
      cls,
      algorithm: str,
      experimenter_factory: experimenters.SerializableExperimenterFactory,
      states: list[benchmarks.BenchmarkState],
      flip_signs_for_min: bool = False,
  ) -> BenchmarkRecord:
    """Generates a BenchmarkRecord from a batch of BenchmarkStates.

    Each state in batch should represent the same study (different repeat).

    Args:
      algorithm: Algorithm name.
      experimenter_factory: Factory used for running BenchmarkState.
      states: List of BenchmarkStates.
      flip_signs_for_min: If true, flip signs of curve when it is MINIMIZE
        metric.

    Returns:
      BenchmarkRecord.
    """
    plot_elements = {}
    plot_elements[RECORD_OBJECTIVE_KEY] = PlotElement(
        curve=cls.to_curve(states, flip_signs_for_min=flip_signs_for_min),
        yscale='symlog',
    )
    return BenchmarkRecord(
        algorithm=algorithm,
        experimenter_metadata=experimenter_factory.dump(),
        plot_elements=plot_elements,
    )


def summarize_element(element: PlotElement) -> np.ndarray:
  if element.plot_type == 'error-bar':
    assert element.curve is not None
    return np.median(np.array(element.curve.ys[:, -1]))
  elif element.plot_type == 'histogram':
    return np.median(element.plot_array)
  else:
    raise NotImplementedError(f'Unsupported plot type {element.plot_type}')


def summarize_elements(
    plot_elements_list: Sequence[PlotElement],
) -> PlotElement:
  summary_vector = [
      summarize_element(element) for element in plot_elements_list
  ]
  return PlotElement(plot_array=np.array(summary_vector), plot_type='histogram')


class BenchmarkRecordAnalyzer:
  """Analyzer for a sequence of Benchmark Records."""

  @classmethod
  def add_comparison_metrics(
      cls,
      records: Sequence[BenchmarkRecord],
      baseline_algo: str,
      *,
      compare_metric: str = RECORD_OBJECTIVE_KEY,
      comparator_factory: convergence_curve.ConvergenceComparatorFactory = convergence_curve.LogEfficiencyConvergenceCurveComparator,
  ) -> list[BenchmarkRecord]:
    """Adds comparison scores as metrics via PlotElements to BenchmarkRecord.

    Comparisons are done for compare_metric with respect to the baseline_algo.

    Args:
      records: Sequence of BenchmarkRecords
      baseline_algo: Baseline algorithm to be compared against.
      compare_metric: Metric of comparison.
      comparator_factory: Comparator used for scoring.

    Returns:
      List of BenchmarkRecords with comparison scores added as metrics.

    Raises:
      ValueError: When baseline_algo cannot be found in records or the metric
      of comparison does not correspond to a curve.
    """
    records_list = [
        (rec.algorithm, json.dumps(dict(rec.experimenter_metadata)), rec)
        for rec in records
    ]
    df = pd.DataFrame(
        records_list, columns=['algorithm', 'experimenter', 'record']
    )

    analyzed_records = []
    for experimenter_key, experimenter_group in df.groupby('experimenter'):
      # Checks and stores the mapping from algorithm to plot_elements
      algo_to_elements_dict = {}
      for algorithm_name, group in experimenter_group.groupby('algorithm'):
        if not group.size:
          continue
        if len(group) != 1:
          output_str = (
              f'Condense {len(group)} records for {algorithm_name} with exptr '
              f' {experimenter_key} before applying comparisons for'
              f' {group}'
          )
          logging.error('%s', output_str)
          continue
        algo_to_elements_dict[algorithm_name] = group.record.iloc[
            0
        ].plot_elements

      # Finds the baseline algorithm and the comparison element.
      if compare_metric not in algo_to_elements_dict[baseline_algo]:
        raise ValueError(
            f'Compare metric {compare_metric} not in baseline {baseline_algo}'
        )
      baseline_element = algo_to_elements_dict[baseline_algo][compare_metric]
      if baseline_element.plot_type != 'error-bar':
        raise ValueError(
            f'No comparison can be done for {compare_metric} since'
            f' plot type is {baseline_element.plot_type}'
        )

      # Attempts to apply comparison and add comparison metrics.
      for algorithm_name, elems_dict in algo_to_elements_dict.items():
        compared_element = elems_dict[compare_metric]
        comparator = comparator_factory(
            baseline_curve=baseline_element.curve,
            compared_curve=compared_element.curve,
        )
        try:
          dict_key = (
              compare_metric + f':{comparator.name}_curve:' + baseline_algo
          )
          elems_dict[dict_key] = PlotElement(
              curve=comparator.curve(),
              plot_type='error-bar',
          )
        except Exception as e:  # pylint: disable=broad-exception-caught
          output_str = (
              f'Skip comparing curve for algo {algorithm_name} with'
              f' {compared_element} \n due to error {e}'
          )
          logging.error('%s', output_str)
        try:
          dict_key = compare_metric + f':{comparator.name}:' + baseline_algo
          elems_dict[dict_key] = PlotElement(
              plot_array=np.asarray([comparator.score()]),
              plot_type='histogram',
          )
        except Exception as e:  # pylint: disable=broad-exception-caught
          output_str = (
              f'Skip comparing score for algo {algorithm_name} with '
              f'  {compared_element} \n due to error {e}'
          )
          logging.error('%s', output_str)
        analyzed_records.append(
            BenchmarkRecord(
                algorithm=algorithm_name,
                experimenter_metadata=vz.Metadata(json.loads(experimenter_key)),  # pytype: disable=wrong-arg-types  # pandas-drop-duplicates-overloads
                plot_elements=elems_dict,
            )
        )

    return analyzed_records

  @classmethod
  def summarize(
      cls,
      records: Sequence[BenchmarkRecord],
      record_to_reduced_keys: Callable[[BenchmarkRecord], tuple[str, str]],
      *,
      summarize_elements_fn: Callable[
          [Sequence[PlotElement]], PlotElement
      ] = summarize_elements,
  ) -> Sequence[BenchmarkRecord]:
    """Summarizes PlotElements to BenchmarkRecord.

    Comparisons are done for compare_metric with respect to the baseline_algo.

    Args:
      records: Sequence of BenchmarkRecords
      record_to_reduced_keys: Takes a BenchmarkRecord and returns the removed
        key as string and the reduced metadata as a json string. For example, if
        metadata = {'experimenter': Sphere, 'dim': 4} and we want to reduce
        across experimenter, then the removed key would be Sphere and the
        reduced metadata would be '{"dim": 4}'.
      summarize_elements_fn: PlotElement summarization function.

    Returns:
      List of summarized BenchmarkRecords.
    """

    records_list = [
        (rec.algorithm, *record_to_reduced_keys(rec), rec) for rec in records
    ]
    df = pd.DataFrame(
        records_list,
        columns=['algorithm', 'removed_key', 'reduced_metadata', 'record'],
    )

    summarized_records = []
    for algorithm_name, group_by_algo in df.groupby('algorithm'):
      if not group_by_algo.size:
        continue
      for reduced_metadata, group_by_reduced_metadata in group_by_algo.groupby(
          'reduced_metadata'
      ):
        if not group_by_reduced_metadata.size:
          continue
        summarized_records.append(
            BenchmarkRecord(
                algorithm=str(algorithm_name),
                experimenter_metadata=vz.Metadata(json.loads(reduced_metadata)),
                plot_elements=cls._summarize_elements_df(
                    group_by_reduced_metadata,
                    summarize_elements_fn=summarize_elements_fn,
                ),
            )
        )
    return summarized_records

  @classmethod
  def _summarize_elements_df(
      cls,
      elements_df: pd.DataFrame,
      summarize_elements_fn: Callable[
          [Sequence[PlotElement]], PlotElement
      ] = summarize_elements,
  ) -> dict[str, PlotElement]:
    """Summarizes PlotElements via removed_key in a dataframe."""
    key_to_elements_list = collections.defaultdict(list)
    for removed_key, group in elements_df.groupby('removed_key'):
      if len(group) > 1:
        raise ValueError(f'More than 1 group after analysis for {removed_key}')
      for key, element in group.record.iloc[0].plot_elements.items():
        key_to_elements_list[key].append(element)

    # Summarize metadata and plot elements
    summarized_plot_elements = {}
    for key, elements_list in key_to_elements_list.items():
      summarized_plot_elements[key] = summarize_elements_fn(elements_list)
    return summarized_plot_elements
